@@ -4,10 +4,10 @@
  *	  prototypes for the reliable BSD-style signal(2) routine.
  *
  *
- * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2011, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $PostgreSQL: pgsql/src/interfaces/libpq/pqsignal.h,v 1.23 2008/01/01 19:46:00 momjian Exp $
+ * src/interfaces/libpq/pqsignal.h
  *
  * NOTES
  *	  This shouldn't be in libpq, but the monitor and some other
@@ -18,8 +18,6 @@
 #ifndef PQSIGNAL_H
 #define PQSIGNAL_H
 
-#include "postgres_fe.h"
-
 typedef void (*pqsigfunc) (int);
 
 extern pqsigfunc pqsignal(int signo, pqsigfunc func);
